/*
 * Copyright (c) Wipro Technologies Ltd, 2003.  All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write the Free Software Foundation, Inc., 59
 * Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 */
/**************************************************************************
 * 
 *    TEST IDENTIFIER	: clock_settime02 
 * 
 *    EXECUTED BY	: root / superuser
 * 
 *    TEST TITLE	: Basic test for clock_settime(2)
 * 
 *    TEST CASE TOTAL	: 1
 * 
 *    AUTHOR		: Aniruddha Marathe <aniruddha.marathe@wipro.com>
 * 
 *    SIGNALS
 * 	Uses SIGUSR1 to pause before test if option set.
 * 	(See the parse_opts(3) man page).
 *
 *    DESCRIPTION
 *     This is a Phase I test for the clock_settime(2) system call.
 *     It is intended to provide a limited exposure of the system call.
 * 
 * 	Setup:
 *	  Setup signal handling.
 *	  Pause for SIGUSR1 if option specified.
 * 
 * 	Test:
 *	 Loop if the proper options are given.
 *	 Set the parameters of timespec struct
 *	 Execute system call
 *	 Check return code, if system call failed (return=-1)
 *		Log the errno and Issue a FAIL message.
 *	 Otherwise, Issue a PASS message.
 * 
 * 	Cleanup:
 * 	  Print errno log and/or timing stats if options given
 * 
 * USAGE:  <for command-line>
 * clock_settime02 [-c n] [-e] [-i n] [-I x] [-P x] [-t] [-p]
 * where:
 * 	-c n : Run n copies simultaneously. 
 *	-e   : Turn on errno logging.
 *	-i n : Execute test n times.
 *	-I x : Execute test for x seconds.
 *	-p   : Pause for SIGUSR1 before starting
 *	-P x : Pause for x seconds between iterations.
 *	-t   : Turn on syscall timing.
 *
 *RESTRICTIONS:
 * None
 *****************************************************************************/

#include "test.h"
#include "usctest.h"
#include <errno.h>
#include <syscall.h>
#include <time.h>

#ifndef __NR_timer_create
#if defined(__i386__)
#define __NR_timer_create 259
#endif
#endif

#ifndef __NR_clock_settime
#if defined(__i386__)
#define __NR_clock_settime (__NR_timer_create + 5)
#elif defined(__ppc__)
#define __NR_clock_settime 245
#elif defined(__powerpc64__)
#define __NR_clock_settime 245
#elif defined(__x86_64__)
#define __NR_clock_settime 227
#endif
#endif

#ifndef __NR_clock_gettime
#if defined(__i386__)
#define __NR_clock_gettime (__NR_timer_create + 6)
#elif defined(__ppc__)
#define __NR_clock_gettime 246
#elif defined(__powerpc64__)
#define __NR_clock_gettime 246
#elif defined(__x86_64__)
#define __NR_clock_gettime 228
#endif
#endif

/* weak symbols. In newer glibc, these symbols should be present. Then
 * it will supersede the definition from this code using _syscall
 */
#pragma weak clock_settime
#pragma weak clock_gettime

static void setup();
static void cleanup();

char *TCID = "clock_settime02";	/* Test program identifier.    */
int TST_TOTAL = 1;		/* Total number of test cases. */
extern int Tst_count;		/* Test Case counter for tst_* routines */
static struct timespec saved;	/* Used to reset the time */

/* register clock_settime and clock_gettime as system calls */
_syscall2(int, clock_settime, clockid_t, which_clock, const struct timespec *,
	       tp);	
_syscall2(int, clock_gettime, clockid_t, which_clock, struct timespec *, tp);	

int
main(int ac, char **av)
{
	int lc;			/* loop counter */
	char *msg;		/* message returned from parse_opts */
	struct timespec spec;	/* Used to specify time for test */

	/* parse standard options */
	if ((msg = parse_opts (ac, av, (option_t *) NULL, NULL)) != 
			(char *) NULL) {
		tst_brkm(TBROK, tst_exit, "OPTION PARSING ERROR - %s", msg);
	}

	/* perform global setup for test */
	setup();

	/* check looping state if -i option given */
	for (lc = 0; TEST_LOOPING(lc); lc++) {

		/* reset Tst_count in case we are looping. */
		Tst_count = 0;

		spec.tv_sec = 1;
		spec.tv_nsec = 0;

		TEST(clock_settime(CLOCK_REALTIME, &spec));

		/* system call not implemented */
		if (TEST_ERRNO == ENOSYS) {
			Tst_count = TST_TOTAL;
			perror("clock_settime");
			tst_brkm(TBROK, cleanup, "");
		}
		if (TEST_RETURN == -1) {

			TEST_ERROR_LOG(TEST_ERRNO);
			tst_resm(TFAIL, "clock_settime(2) Failed and set errno"
					" to %d", TEST_ERRNO);
		} else {
			tst_resm(TPASS, "clock_settime(2) Passed");
		}
	}		/* End for TEST_LOOPING */

	/* Clean up and exit */
	cleanup();

	/* NOTREACHED */ 
	return 0;
}

/* setup() - performs all ONE TIME setup for this test */
void
setup()
{
	/* capture signals */
	tst_sig(NOFORK, DEF_HANDLER, cleanup);

	/* Check whether we are root */
	if (geteuid() != 0) {
		tst_brkm(TBROK, tst_exit, "Test must be run as root");
	}
	/* Save the current time specifications */
	if (clock_gettime(CLOCK_REALTIME, &saved) < 0) {
		if (errno == ENOSYS) {
			/* System call not implemened */
			perror("clock_gettime");
			tst_brkm(TBROK, tst_exit, "");
		}
		tst_brkm(TBROK, tst_exit, "Could not save the current time");
	}

	/* Pause if that option was specified */
	TEST_PAUSE;
}	/* End setup() */

/*
 * cleanup() - Performs one time cleanup for this test at
 * completion or premature exit
 */

void
cleanup()
{
	/* Set the saved time */
	if (clock_settime(CLOCK_REALTIME, &saved) < 0) {
		tst_resm(TWARN, "FATAL COULD NOT RESET THE CLOCK");
		tst_resm(TFAIL, "Error Setting Time, errno=%d", errno);
	}

	/*
	* print timing stats if that option was specified.
	* print errno log if that option was specified.
	*/
	TEST_CLEANUP;

	/* exit with return code appropriate for results */
	tst_exit();
}	/* End cleanup() */
