/*
 * Copyright (c) Wipro Technologies Ltd, 2003.  All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write the Free Software Foundation, Inc., 59
 * Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 */
/**************************************************************************
 * 
 *    TEST IDENTIFIER	: timer_delete03 
 * 
 *    EXECUTED BY	: anyone
 * 
 *    TEST TITLE	: Test checking for basic error conditions for
 *    			  timer_delete(2)
 * 
 *    TEST CASE TOTAL	: 1
 * 
 *    AUTHOR		: Aniruddha Marathe <aniruddha.marathe@wipro.com>
 * 
 *    SIGNALS
 * 	Uses SIGUSR1 to pause before test if option set.
 * 	(See the parse_opts(3) man page).
 *
 *    DESCRIPTION
 *    	This test case check whether timer_delete(2) returns appropriate error
 *    	value for invalid parameter
 *
 * 	Setup:
 *	 Setup signal handling.
 *	 Pause for SIGUSR1 if option specified.
 * 
 * 	Test:
 *	 Loop if the proper options are given.
 *	 Execute system call with invalid parameter.
 *	 Check return code, if system call fails with errno == expected errno
 * 	 	Issue syscall passed with expected errno
 *	 Otherwise, Issue syscall failed to produce expected errno
 * 
 * 	Cleanup:
 * 	 Print errno log and/or timing stats if options given
 * 
 * USAGE:  <for command-line>
 * timer_delete03 [-c n] [-e] [-i n] [-I x] [-P x] [-t] [-p]
 * where:
 * 	-c n : run n copies simultaneously
 *	-e   : Turn on errno logging.
 *	-i n : Execute test n times.
 *	-I x : Execute test for x seconds.
 *	-p   : Pause for SIGUSR1 before starting
 *	-P x : Pause for x seconds between iterations.
 *	-t   : Turn on syscall timing.
 *
 * RESTRICTIONS:
 * None
 *****************************************************************************/

#include "test.h"
#include "usctest.h"
#include <errno.h>
#include <syscall.h>
#include <time.h>
#include <signal.h>

#ifndef __NR_timer_create
#if defined(__i386__)
#define __NR_timer_create 259
#endif
#endif

#ifndef __NR_timer_delete
#if defined(__i386__)
#define __NR_timer_delete (__NR_timer_create + 4)
#elif defined(__ppc__)
#define __NR_timer_delete 244
#elif defined(__powerpc64__)
#define __NR_timer_delete 244
#elif defined(__x86_64__)
#define __NR_timer_delete 226
#endif
#endif

#define INVALID_ID -1

/* weak symbol. In newer glibc, there should be a definition of timer_delete
 * at that time, it will supersede the definition in this code
 */ 
#pragma weak timer_delete

static void setup();
static void cleanup();

char *TCID = "timer_delete03"; 	/* Test program identifier.    */
int TST_TOTAL;		       	/* Total number of test cases. */
extern int Tst_count;    	/* Test Case counter for tst_* routines */

static int exp_enos[] = {EINVAL, 0};

static struct test_case_t {
	char *err_desc;		/* error description */
	int  exp_errno;		/* expected error number*/
	char *exp_errval;	/* Expected errorvalue string*/
} testcase[] = {
	{"Invalid parameter", EINVAL, "EINVAL"} /* Invalid timer ID */
};

/* register timer_delete as system call */
_syscall1(int, timer_delete, timer_t, timer_id);

int
main(int ac, char **av)
{
	int lc, i;	/* loop counter */
	char *msg;	/* message returned from parse_opts */

	/* parse standard options */
	if ((msg = parse_opts(ac, av, (option_t *)NULL, NULL))
		!= (char *)NULL) {
		tst_brkm(TBROK, tst_exit, "OPTION PARSING ERROR - %s", msg);
	}

	TST_TOTAL = sizeof(testcase) / sizeof(testcase[0]);

	/* perform global setup for test */
	setup();

	/* check looping state if -i option given */
	for (lc = 0; TEST_LOOPING(lc); lc++) {

		/* reset Tst_count in case we are looping. */
		Tst_count = 0;

		for (i = 0; i < TST_TOTAL; i++) {

			TEST(timer_delete(INVALID_ID));

			if (TEST_ERRNO == ENOSYS) {
				/* System call is not implemented
				 * In the running kernel
				 */
				Tst_count = TST_TOTAL;
				perror("timer_delete");
				tst_brkm(TBROK, cleanup, "");
			}
			/* check return code */
			if ((TEST_RETURN == -1) && (TEST_ERRNO == testcase[i].
						exp_errno)) {
				tst_resm(TPASS, "timer_delete(2) expected"
						" failure; Got errno - %s : %s"
						, testcase[i].exp_errval,
						testcase[i].err_desc);
			} else {
				tst_resm(TFAIL, "timer_delete(2) failed to"
						" produce expected error; %d"
						" , errno : %s and got %d",
						testcase[i].exp_errno,
						testcase[i].exp_errval,
						TEST_ERRNO);
			} /* end of else */

			TEST_ERROR_LOG(TEST_ERRNO);
		}	/* End of TEST CASE LOOPING*/
	}	/* End for TEST_LOOPING*/

	/* Clean up and exit */
	cleanup();

	/* NOTREACHED */
	return 0;
}

/* setup() - performs all ONE TIME setup for this test */
void
setup()
{
	/* capture signals */
	tst_sig(NOFORK, DEF_HANDLER, cleanup);

	/* set the expected errnos... */
	TEST_EXP_ENOS(exp_enos);

	/* Pause if that option was specified */
	TEST_PAUSE;
}	/* End setup() */

/*
 * cleanup() - Performs one time cleanup for this test at
 * completion or premature exit
 */

void
cleanup()
{
	/*
	* print timing stats if that option was specified.
	* print errno log if that option was specified.
	*/
	TEST_CLEANUP;

	/* exit with return code appropriate for results */
	tst_exit();
}	/* End cleanup() */
